import os
import tempfile
import shutil
from xml.dom import minidom

from docx_charts.chart import Chart


class Document:
	'''
	Represents a Word document.

	Attributes:
		file_path (str): The path to the Word document.
		extracted (tempfile.TemporaryDirectory): A temporary directory containing the extracted contents of the Word document.
	'''
	file_path: str
	extracted: tempfile.TemporaryDirectory


	def __init__(self, file_path: str):
		'''
		Initializes a new instance of Document.

		Args:
			file_path (str): The path to the Word document file.
		'''
		self.file_path = file_path
		self.extracted = tempfile.TemporaryDirectory()
		shutil.unpack_archive(file_path, self.extracted.name, format='zip')


	def __del__(self):
		'''
		Cleans up the temporary directory when the Document object is destroyed.
		'''
		self.extracted.cleanup()


	def save(self, file_path: str|None = None):
		'''
		Re-archives the Word document and saves it to the specified path.

		Args:
			file_path (str): The path to save the Word document to. If not specified, the original file will be overwritten.
		'''
		if file_path is None:
			file_path = self.file_path
		shutil.make_archive(file_path, 'zip', self.extracted.name)
		shutil.move(f'{file_path}.zip', file_path)


	def list_contents(self):
		'''
		Lists the contents of the Word document.

		Returns:
			A list of files in the root of the extracted archive
		'''
		return os.listdir(self.extracted.name)


	def charts(self) -> list[Chart]:
		'''
		Lists the charts in the Word document.

		Returns:
			A list of objects representing the charts in the Word document.
		'''
		charts: list[Chart] = []
		with open(os.path.join(self.extracted.name, 'word/document.xml'), encoding='utf-8') as doc:
			with open(os.path.join(self.extracted.name, 'word/_rels/document.xml.rels'), encoding='utf-8') as rels:
				doc_dom = minidom.parse(doc)
				rels_dom = minidom.parse(rels)
				for node in doc_dom.getElementsByTagName('c:chart'):
					relationship_id = node.getAttribute('r:id')
					relationship = [rel for rel in rels_dom.getElementsByTagName('Relationship') if rel.getAttribute('Id') == relationship_id][0]
					path = os.path.join(self.extracted.name, 'word', relationship.getAttribute('Target'))
					name = node.parentNode.parentNode.parentNode.getElementsByTagName('wp:docPr')[0].getAttribute('name')
					charts.append(Chart(path, name))
		return charts


	def charts_by_name(self, name: str) -> list[Chart]:
		'''
		Finds the charts in the Word document with the specified name.

		Args:
			name (str): The name of the charts to find.

		Returns:
			A list of objects representing the charts in the Word document with the specified name.

		Note:
			Generally this will only return one chart, but technically it's possible for there to be multiple charts with the same name.
		'''
		return [chart for chart in self.charts() if chart.name == name]



import io
from xml.dom import minidom


Series = dict[str, float|None]  # category: value
ChartData = dict[str, Series]  # series_name: Series


def series_name(series: minidom.Element) -> str | None:
	if not series.getElementsByTagName('c:tx'):
		return None
	return series.getElementsByTagName('c:tx')[0].getElementsByTagName('c:v')[0].firstChild.nodeValue  # type: ignore


class Chart:
	'''
	Represents a chart in a Word document.

	Attributes:
		file (io.TextIOWrapper): The file handler for the chart's XML.
		name (str): The name of the chart.
	'''
	file: io.TextIOWrapper
	name: str


	def __init__(self, path: str, name: str):
		'''
		Initializes a new instance of Chart.

		Args:
			path (str): The path to the chart's XML file.
			name (str): The name of the chart.
		'''
		self.file = open(path, 'r+', encoding='utf-8')
		self.name = name


	def __del__(self):
		'''
		Closes the file handler when the Chart object is destroyed.
		'''
		self.file.close()


	def __str__(self) -> str:
		'''
		Returns a string representation of the Chart object.

		Returns:
			A string representation of the Chart object.
		'''
		return f'{self.name} ({self.file.name})'

	__repr__ = __str__


	def data(self) -> ChartData:
		'''
		Gets the internal table data the chart is based on.

		Returns:
			A dictionary of Series dicts representing data for the chart.
		'''
		dom = minidom.parse(self.file)
		series: ChartData = {
			series_name(series) or f'Series{i+1}':
			dict(zip(
				[cat.firstChild.nodeValue for cat in series.getElementsByTagName('c:cat')[0].getElementsByTagName('c:v')
					if cat.firstChild and isinstance(cat.firstChild, minidom.Text) and isinstance(cat.firstChild.nodeValue, str)],
				[float(val.firstChild.nodeValue) for val in series.getElementsByTagName('c:val')[0].getElementsByTagName('c:v')
					if val.firstChild and isinstance(val.firstChild, minidom.Text)]
			))
			for i, series in enumerate(dom.getElementsByTagName('c:ser'))
		}
		self.file.seek(0)
		return series


	def write(self, data: ChartData) -> None:
		'''
		Overwrites the data of the chart.

		Args:
			data (dict[str, dict[str, float|None]]): The data to write to the chart. The first key is the name of the series, the second key is the name of the category.

		Note:
			The names of the series and categories must be the same as in the original data.
			However, unchanged series and categories may be left out.
			If a value is None, it will be removed from the chart.
		'''
		dom = minidom.parse(self.file)

		for (new_series_name, new_series) in data.items():
			try:
				series = [series for i, series in enumerate(dom.getElementsByTagName('c:ser')) if (series_name(series) or f'Series{i+1}') == new_series_name][0]
			except IndexError:
				raise ValueError(f'Chart "{self.name}" does not contain a series named "{new_series_name}".')
			cats = [cat.firstChild for cat in series.getElementsByTagName('c:cat')[0].getElementsByTagName('c:v')
				if cat.firstChild and isinstance(cat.firstChild, minidom.Text)]
			vals = [val.firstChild for val in series.getElementsByTagName('c:val')[0].getElementsByTagName('c:v')
				if val.firstChild and isinstance(val.firstChild, minidom.Text)]

			# For every value in the new data, find the old value in the chart and overwrite it.
			for new_cat, new_val in new_series.items():
				if new_cat not in [cat.nodeValue for cat in cats]:
					raise ValueError(f'Chart "{self.name}" does not contain a category named "{new_cat}".')
				for cat, val in zip(cats, vals):
					if cat.nodeValue == new_cat:
						# Overwrite the value in the chart. Or if it is None, remove it from the chart.
						if new_val is None:
							val.parentNode.removeChild(val)
						else:
							val.replaceWholeText(str(new_val))
						break

		self.file.seek(0)
		self.file.truncate()
		self.file.write(dom.toxml())
		self.file.flush()
		self.file.seek(0)
